import numpy as np
import torch
import torchmetrics
from rich.console import Console
from rich.table import Table


def calculate_retrieval_metrics(
    labels: np.ndarray, retrieved_ids: np.ndarray, top_k: int
) -> dict:
    """
    Calculate retrieval metrics given labels and retrieved IDs

    Args:
        labels: Ground truth labels for each query
        retrieved_ids: Retrieved IDs for each query
        top_k: Number of top results retrieved

    Returns:
        dict: Dictionary containing metric names and scores
    """
    matches = np.expand_dims(labels, axis=1) == labels[retrieved_ids]
    matches = torch.tensor(np.array(matches), dtype=torch.float16)
    targets = torch.ones(matches.shape)
    indexes = (
        torch.arange(matches.shape[0]).view(-1, 1)
        * torch.ones(1, matches.shape[1]).long()
    )

    metrics = [
        torchmetrics.retrieval.RetrievalMRR(),
        torchmetrics.retrieval.RetrievalNormalizedDCG(),
        torchmetrics.retrieval.RetrievalPrecision(),
        torchmetrics.retrieval.RetrievalRecall(),
        torchmetrics.retrieval.RetrievalHitRate(),
        torchmetrics.retrieval.RetrievalMAP(),
    ]
    eval_metrics_results = {}

    for metr in metrics:
        score = round(metr(targets, matches, indexes).item(), 4)
        metr_name = metr.__class__.__name__.replace("Retrieval", "")
        eval_metrics_results[metr_name] = score

    # Print metrics in a rich table
    table = Table(title=f"Retrieval Metrics @ k={top_k}")
    table.add_column("Metric", style="cyan")
    table.add_column("Score", style="magenta")

    for metric_name, score in eval_metrics_results.items():
        table.add_row(metric_name, f"{score:.4f}")

    console = Console()
    console.print(table)

    return eval_metrics_results



import bm25s
import numpy as np
import Stemmer

from xretrieval.models_registry import ModelRegistry


@ModelRegistry.register("xhluca/bm25s", model_input="text")
class BM25sModel:
    def __init__(self, model_id: str = "xhluca/bm25s"):
        self.model_id = model_id
        self.model = self.load_model()

        self.corpus_tokens = None
        self.stemmer = Stemmer.Stemmer("english")

    def load_model(self):
        return bm25s.BM25()

    def tokenize_text(self, text: list[str]):
        corpus_tokens = bm25s.tokenize(text, stopwords="en", stemmer=self.stemmer)
        self.model.index(corpus_tokens)
        self.corpus_tokens = corpus_tokens

    def retrieve(self, queries: list[str], top_k: int) -> np.ndarray:
        queries_tokens = bm25s.tokenize(queries, stopwords="en", stemmer=self.stemmer)
        results = self.model.retrieve(
            queries_tokens, k=top_k + 1
        )  # +1 for self-matches

        retrieved_ids = []
        # Filter self matches for each query
        for idx, docs in enumerate(results.documents):
            filtered_docs = [doc for doc in docs if doc != idx][:top_k]
            retrieved_ids.append(filtered_docs)
        retrieved_ids = np.array(retrieved_ids)

        return retrieved_ids



from .blip2 import BLIP2ImageModel, BLIP2Model, BLIP2TextModel
from .bm25 import BM25sModel
from .sentence_transformers import SentenceTransformerModel
from .timm import TimmModel

__all__ = [
    "SentenceTransformerModel",
    "TimmModel",
    "BLIP2Model",
    "BLIP2TextModel",
    "BLIP2ImageModel",
    "BM25sModel",
]



__version__ = "0.2.0"
__author__ = "Dickson Neoh"
__email__ = "dickson.neoh@gmail.com"

from .core import (
    list_datasets,
    list_models,
    load_dataset,
    load_model,
    run_benchmark,
    run_benchmark_bm25,
    visualize_ground_truth,
    visualize_retrieval,
)
from .datasets import COCODataset, COCODatasetBLIP2Captions, COCODatasetVLRMCaptions
from .datasets_registry import DatasetRegistry
from .models import SentenceTransformerModel
from .models_registry import ModelRegistry
from .rrf import run_rrf

__all__ = [
    "list_datasets",
    "load_dataset",
    "list_models",
    "load_model",
    "run_benchmark",
    "run_benchmark_bm25",
    "run_rrf",
    "visualize_retrieval",
    "visualize_ground_truth",
]



from dataclasses import dataclass
from typing import Any, Callable, Literal, Type

# Type aliases for clarity
ModelType = Type[Any]  # Type of the model class
DecoratorFunction = Callable[[ModelType], ModelType]


@dataclass
class ModelInfo:
    """Stores information about a registered model."""

    model_id: str
    model_class: ModelType
    model_input: Literal["text", "image", "multi-modal"]


class ModelRegistry:
    """Registry for managing Sentence Transformer models."""

    _models: dict[str, ModelInfo] = {}

    @classmethod
    def register(
        cls, model_id: str, model_input: Literal["text", "image", "text-image"]
    ) -> DecoratorFunction:
        """Decorator to register a model.

        Args:
            model_id: HuggingFace model ID for the Sentence Transformer

        Returns:
            Decorator function that registers the model class
        """

        def decorator(model_class: ModelType) -> ModelType:
            if model_id in cls._models:
                raise ValueError(f"Model '{model_id}' is already registered")

            cls._models[model_id] = ModelInfo(
                model_id=model_id,
                model_class=model_class,
                model_input=model_input,
            )
            return model_class

        return decorator

    @classmethod
    def get(cls, model_id: str) -> ModelType:
        """Retrieve a model class by name."""
        if model_id not in cls._models:
            raise KeyError(f"Model '{model_id}' not found in registry")

        model_info = cls._models[model_id]
        return model_info.model_class

    @classmethod
    def get_model_info(cls, model_id: str) -> ModelInfo:
        """Retrieve a model info by name."""
        if model_id not in cls._models:
            raise KeyError(f"Model '{model_id}' not found in registry")

        return cls._models[model_id]

    @classmethod
    def list(cls) -> dict[str, str]:
        """List all registered models with their types.

        Returns:
            A dictionary mapping model IDs to their types (text/image/multi-modal)
        """
        return {model_id: info.model_input for model_id, info in cls._models.items()}



from dataclasses import dataclass
from typing import Any, Callable, Type

# Type aliases for clarity
DatasetType = Type[Any]  # Type of the dataset class
DecoratorFunction = Callable[[DatasetType], DatasetType]


@dataclass
class DatasetInfo:
    """Stores information about a registered dataset."""

    name: str
    dataset_class: DatasetType
    description: str = ""


class DatasetRegistry:
    """Registry for managing datasets."""

    _datasets: dict[str, DatasetInfo] = {}

    @classmethod
    def register(cls, name: str, description: str = "") -> DecoratorFunction:
        """Decorator to register a dataset.

        Args:
            name: Name for the dataset
            description: Optional description of the dataset

        Returns:
            Decorator function that registers the dataset class
        """

        def decorator(dataset_class: DatasetType) -> DatasetType:
            if name in cls._datasets:
                raise ValueError(f"Dataset '{name}' is already registered")

            cls._datasets[name] = DatasetInfo(
                name=name, dataset_class=dataset_class, description=description
            )
            return dataset_class

        return decorator

    @classmethod
    def get(cls, name: str) -> Any:
        """Retrieve a dataset instance by name."""
        if name not in cls._datasets:
            raise KeyError(f"Dataset '{name}' not found in registry")
        return cls._datasets[name].dataset_class()

    @classmethod
    def list(cls) -> dict[str, DatasetInfo]:
        """List all registered datasets with their descriptions."""
        return {name: info.description for name, info in cls._datasets.items()}



import json
from pathlib import Path

import pandas as pd
from loguru import logger

from ..datasets_registry import DatasetRegistry


@DatasetRegistry.register("coco-val-2017", "The COCO Validation Set with 5k images.")
class COCODataset:
    def __init__(
        self,
        data_dir: str = "./data/coco/",
    ):
        self.data_dir = Path(data_dir)
        self.annotations_dir = self.data_dir / "annotations"
        self.images_dir = self.data_dir / "val2017"

        # Check if dataset exists, download if not
        if not self.images_dir.exists():
            logger.info(f"COCO validation dataset not found in {self.images_dir}")
            logger.info("Downloading COCO validation dataset...")
            self.download()
        else:
            logger.info(
                f"COCO validation dataset found in {self.images_dir}, skipping download"
            )

    def download(self):
        """Download and extract COCO validation dataset and annotations if not already present."""
        import os
        import zipfile

        import requests
        from tqdm import tqdm

        def download_file(url, filename):
            """Helper function to download a file with progress bar."""
            response = requests.get(url, stream=True)
            total_size = int(response.headers.get("content-length", 0))

            with open(filename, "wb") as file, tqdm(
                desc=f"Downloading {filename}",
                total=total_size,
                unit="iB",
                unit_scale=True,
                unit_divisor=1024,
            ) as pbar:
                for data in response.iter_content(chunk_size=1024):
                    size = file.write(data)
                    pbar.update(size)

        # Store the original working directory
        original_dir = os.getcwd()

        try:
            # Create directories if they don't exist
            os.makedirs(self.data_dir, exist_ok=True)
            os.makedirs(self.data_dir / "annotations", exist_ok=True)

            # Change to the coco directory
            os.chdir(self.data_dir)

            # Download and extract validation images
            logger.info("Downloading COCO validation dataset...")
            download_file(
                "http://images.cocodataset.org/zips/val2017.zip", "val2017.zip"
            )

            logger.info("Extracting images...")
            with zipfile.ZipFile("val2017.zip", "r") as zip_ref:
                zip_ref.extractall()
            os.remove("val2017.zip")

            # Download and extract annotations
            logger.info("Downloading COCO annotations...")
            download_file(
                "http://images.cocodataset.org/annotations/annotations_trainval2017.zip",
                "annotations_trainval2017.zip",
            )

            logger.info("Extracting annotations...")
            with zipfile.ZipFile("annotations_trainval2017.zip", "r") as zip_ref:
                zip_ref.extractall()
            os.remove("annotations_trainval2017.zip")

            logger.info("Download and extraction complete!")

        finally:
            # Always restore the original working directory
            os.chdir(original_dir)

    def load_annotations(self) -> pd.DataFrame:
        """Load and process COCO annotations."""
        # Load caption and instance annotations
        captions_file = self.annotations_dir / "captions_val2017.json"
        instances_file = self.annotations_dir / "instances_val2017.json"

        with open(captions_file, "r") as f:
            coco_captions = json.load(f)
        with open(instances_file, "r") as f:
            coco_instances = json.load(f)

        # Create DataFrames
        df_images = pd.DataFrame(coco_captions["images"])
        df_captions = pd.DataFrame(coco_captions["annotations"])
        df_instances = pd.DataFrame(coco_instances["annotations"])
        df_categories = pd.DataFrame(coco_instances["categories"])

        # Prepare category information
        df_categories = df_categories.rename(columns={"id": "category_id"})
        df_instances = df_instances[["image_id", "category_id"]]

        # Get categories per image
        df_image_categories = pd.merge(
            df_instances, df_categories, on="category_id", how="left"
        )

        # Group categories by image
        df_image_categories = (
            df_image_categories.groupby("image_id")["name"]
            .agg(lambda x: ",".join(sorted(list(set(x)))))
            .reset_index()
        )

        # Merge everything together
        df_images = df_images.rename(columns={"id": "image_id"})
        df = pd.merge(
            df_images[["file_name", "image_id"]],
            df_captions[["image_id", "caption"]],
            how="left",
            on="image_id",
        )

        df = pd.merge(df, df_image_categories, how="left", on="image_id")

        # Add full image paths
        df["image_path"] = df.file_name.apply(lambda x: str(self.images_dir / x))

        return df

    def get_dataset(self) -> pd.DataFrame:
        """Get the processed COCO dataset."""
        df = self.load_annotations()

        # Group multiple captions per image into a single row
        df = (
            df.groupby("image_id")
            .agg(
                {
                    "file_name": "first",
                    "image_path": "first",
                    "caption": lambda x: " ".join(x),
                    "name": "first",
                }
            )
            .reset_index()
        )

        return df


@DatasetRegistry.register(
    "coco-val-2017-blip2-captions",
    "The COCO Validation Set with 5k images and BLIP2 captions.",
)
class COCODatasetBLIP2Captions(COCODataset):
    def load_annotations(self) -> pd.DataFrame:
        url = "https://github.com/dnth/x.retrieval/releases/download/v0.1.1/blip2_captioned_coco_val_2017.parquet"
        df = pd.read_parquet(url)
        return df


@DatasetRegistry.register(
    "coco-val-2017-vlrm-captions",
    "The COCO Validation Set with 5k images and VLRM captions.",
)
class COCODatasetVLRMCaptions(COCODataset):
    def load_annotations(self) -> pd.DataFrame:
        url = "https://github.com/dnth/x.retrieval/releases/download/v0.1.1/vlrm_captioned_coco_val_2017.parquet"
        df = pd.read_parquet(url)
        return df



import numpy as np
import pandas as pd
import torch
import torchmetrics
from rich.console import Console
from rich.table import Table

from .core import load_dataset


def run_rrf(
    results_list: list[pd.DataFrame], dataset: str, top_k: int = 10
) -> pd.DataFrame:
    """
    Combines multiple retrieval results using Reciprocal Rank Fusion algorithm.

    Args:
        results_list: List of DataFrames containing retrieval results
        dataset: Dataset supported by xretrieval. Run `xretrieval.list_datasets()` to see available datasets.
        top_k: Number of top results to retrieve

    Returns:
        DataFrame with combined retrieval results
    """

    dataset = load_dataset(dataset)

    # Initialize lists for all columns
    new_retrieved_ids = []
    new_retrieved_paths = []
    new_retrieved_captions = []
    new_retrieved_names = []
    new_is_correct = []

    # Get retrieved IDs from all results
    retrieved_ids_lists = [df["retrieved_ids"].tolist() for df in results_list]

    # Iterate through each query
    for idx in range(len(results_list[0])):
        # Get rankings for current query from all results
        rankings = [results[idx] for results in retrieved_ids_lists]

        # Apply RRF to get sorted doc IDs and limit to top_k
        rrf_scores = reciprocal_rank_fusion(rankings)
        sorted_docs = [
            doc_id
            for doc_id, _ in sorted(
                rrf_scores.items(), key=lambda x: x[1], reverse=True
            )
        ][:top_k]  # Limit to top_k results

        # Get corresponding values from dataset
        paths = [
            dataset[dataset["image_id"] == doc_id]["image_path"].iloc[0]
            for doc_id in sorted_docs
        ]
        captions = [
            dataset[dataset["image_id"] == doc_id]["caption"].iloc[0]
            for doc_id in sorted_docs
        ]
        names = [
            dataset[dataset["image_id"] == doc_id]["name"].iloc[0]
            for doc_id in sorted_docs
        ]

        # Check if retrieved IDs contain the query ID
        query_id = results_list[0].iloc[idx]["query_id"]
        is_correct = [doc_id == query_id for doc_id in sorted_docs]

        # Append to lists
        new_retrieved_ids.append(sorted_docs)
        new_retrieved_paths.append(paths)
        new_retrieved_captions.append(captions)
        new_retrieved_names.append(names)
        new_is_correct.append(is_correct)

    # Create new dataframe with updated columns
    new_df = results_list[0].copy()
    new_df["retrieved_ids"] = new_retrieved_ids
    new_df["retrieved_paths"] = new_retrieved_paths
    new_df["retrieved_captions"] = new_retrieved_captions
    new_df["retrieved_names"] = new_retrieved_names

    # Recalculate is_correct based on query_name matching retrieved_names
    new_df["is_correct"] = new_df.apply(
        lambda row: [name == row["query_name"] for name in row["retrieved_names"]],
        axis=1,
    )

    matches = np.array(new_df["is_correct"].tolist())
    matches = torch.tensor(matches, dtype=torch.float16)
    targets = torch.ones(matches.shape)
    indexes = (
        torch.arange(matches.shape[0]).view(-1, 1)
        * torch.ones(1, matches.shape[1]).long()
    )

    metrics = [
        torchmetrics.retrieval.RetrievalMRR(),
        torchmetrics.retrieval.RetrievalNormalizedDCG(),
        torchmetrics.retrieval.RetrievalPrecision(),
        torchmetrics.retrieval.RetrievalRecall(),
        torchmetrics.retrieval.RetrievalHitRate(),
        torchmetrics.retrieval.RetrievalMAP(),
    ]
    eval_metrics_results = {}

    for metr in metrics:
        score = round(metr(targets, matches, indexes).item(), 4)
        metr_name = metr.__class__.__name__.replace("Retrieval", "")
        eval_metrics_results[metr_name] = score

    # Print metrics in a rich table
    table = Table(title=f"Retrieval Metrics @ k={top_k}")
    table.add_column("Metric", style="cyan")
    table.add_column("Score", style="magenta")

    for metric_name, score in eval_metrics_results.items():
        table.add_row(metric_name, f"{score:.4f}")

    console = Console()
    console.print(table)

    return eval_metrics_results, new_df


def reciprocal_rank_fusion(ranked_lists: list[list], bias: int = 60) -> dict:
    """
    Combines multiple ranked lists using Reciprocal Rank Fusion algorithm.

    Args:
        ranked_lists: List of lists, where each sublist contains document IDs in ranked order
        bias: Constant that smooths the impact of high rankings (default: 60)

    Returns:
        Dictionary mapping document IDs to their combined RRF scores, sorted by score
    """
    fusion_scores = {}

    # Calculate RRF score for each document in each ranking
    for ranked_list in ranked_lists:
        for position, document_id in enumerate(ranked_list, start=1):
            if document_id not in fusion_scores:
                fusion_scores[document_id] = 0

            # RRF formula: 1 / (rank + bias)
            rrf_score = 1 / (position + bias)
            fusion_scores[document_id] += rrf_score

    # Sort documents by their fusion scores in descending order
    sorted_results = dict(
        sorted(fusion_scores.items(), key=lambda item: item[1], reverse=True)
    )

    return sorted_results



import numpy as np
import torch
from PIL import Image
from tqdm.auto import tqdm
from transformers import (
    AutoProcessor,
    Blip2TextModelWithProjection,
    Blip2VisionModelWithProjection,
)

from xretrieval.models.base import ImageModel, MultiModalModel, TextModel
from xretrieval.models_registry import ModelRegistry


@ModelRegistry.register(
    "transformers/Salesforce/blip2-itm-vit-g",
    model_input="text-image",
)
class BLIP2Model(MultiModalModel):
    def __init__(self, model_id: str):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model_id = model_id.replace("transformers/", "")
        self.text_model, self.vision_model, self.processor = self.load_models()
        self.text_model.to(self.device)
        self.vision_model.to(self.device)
        self.text_model.eval()
        self.vision_model.eval()

    def load_models(self):
        text_model = Blip2TextModelWithProjection.from_pretrained(
            self.model_id, torch_dtype=torch.float16, device_map="auto"
        )
        vision_model = Blip2VisionModelWithProjection.from_pretrained(
            self.model_id, torch_dtype=torch.float16, device_map="auto"
        )
        processor = AutoProcessor.from_pretrained(self.model_id)
        return text_model, vision_model, processor

    def encode_text(self, captions: list[str], batch_size: int = 32) -> np.ndarray:
        all_features = []
        for i in tqdm(range(0, len(captions), batch_size), desc="Encoding captions"):
            batch_captions = captions[i : i + batch_size]
            inputs = self.processor(
                text=batch_captions, return_tensors="pt", padding=True
            ).to(self.device)

            with torch.inference_mode():
                text_features = (
                    self.text_model(**inputs)
                    .text_embeds.detach()
                    .cpu()
                    .numpy()
                    .astype(np.float32)
                )
            all_features.append(text_features[:, 0, :])
        return np.concatenate(all_features, axis=0)

    def encode_image(self, image_paths: list[str], batch_size: int = 32) -> np.ndarray:
        all_features = []
        for i in tqdm(range(0, len(image_paths), batch_size), desc="Encoding images"):
            batch_paths = image_paths[i : i + batch_size]
            images = [Image.open(path).convert("RGB") for path in batch_paths]
            inputs = self.processor(
                images=images, return_tensors="pt", padding=True
            ).to(self.device)

            with torch.inference_mode():
                image_features = (
                    self.vision_model(**inputs)
                    .image_embeds.detach()
                    .cpu()
                    .numpy()
                    .astype(np.float32)
                )
            all_features.append(image_features[:, 0, :])
        return np.concatenate(all_features, axis=0)


@ModelRegistry.register(
    "transformers/Salesforce/blip2-itm-vit-g-text",
    model_input="text",
)
class BLIP2TextModel(TextModel):
    def __init__(self, model_id: str):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model_id = model_id.replace("transformers/", "")
        self.model_id = self.model_id.replace("-text", "")
        self.model, self.processor = self.load_model()
        self.model.to(self.device)
        self.model.eval()

    def load_model(self):
        model = Blip2TextModelWithProjection.from_pretrained(
            self.model_id, torch_dtype=torch.float16, device_map="auto"
        )
        processor = AutoProcessor.from_pretrained(self.model_id)
        return model, processor

    def encode_text(self, captions: list[str], batch_size: int = 32) -> np.ndarray:
        all_features = []

        for i in tqdm(range(0, len(captions), batch_size), desc="Encoding captions"):
            batch_captions = captions[i : i + batch_size]
            inputs = self.processor(
                text=batch_captions, return_tensors="pt", padding=True
            ).to(self.device)

            with torch.inference_mode():
                text_features = (
                    self.model(**inputs)
                    .text_embeds.detach()
                    .cpu()
                    .numpy()
                    .astype(np.float32)
                )

            all_features.append(text_features[:, 0, :])

        return np.concatenate(all_features, axis=0)


@ModelRegistry.register(
    "transformers/Salesforce/blip2-itm-vit-g-image",
    model_input="image",
)
class BLIP2ImageModel(ImageModel):
    def __init__(self, model_id: str):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model_id = model_id.replace("transformers/", "")
        self.model_id = self.model_id.replace("-image", "")
        self.model, self.processor = self.load_model()
        self.model.to(self.device)
        self.model.eval()

    def load_model(self):
        model = Blip2VisionModelWithProjection.from_pretrained(
            self.model_id, torch_dtype=torch.float16, device_map="auto"
        )
        processor = AutoProcessor.from_pretrained(self.model_id)
        return model, processor

    def encode_image(self, image_paths: list[str], batch_size: int = 32) -> np.ndarray:
        all_features = []
        for i in tqdm(range(0, len(image_paths), batch_size), desc="Encoding images"):
            batch_paths = image_paths[i : i + batch_size]
            images = [Image.open(path).convert("RGB") for path in batch_paths]
            inputs = self.processor(
                images=images, return_tensors="pt", padding=True
            ).to(self.device)

            with torch.inference_mode():
                image_features = (
                    self.model(**inputs)
                    .image_embeds.detach()
                    .cpu()
                    .numpy()
                    .astype(np.float32)
                )

            all_features.append(image_features[:, 0, :])

        return np.concatenate(all_features, axis=0)



import numpy as np
import timm
import torch
from PIL import Image
from tqdm.auto import tqdm

from xretrieval.models.base import ImageModel
from xretrieval.models_registry import ModelRegistry


@ModelRegistry.register("timm/resnet18.a1_in1k", model_input="image")
class TimmModel(ImageModel):
    def __init__(self, model_id: str):
        self.model_id = model_id
        self.model = self.load_model()
        self.model.eval()
        self.transform = timm.data.create_transform(
            **timm.data.resolve_model_data_config(self.model)
        )

    def load_model(self):
        return timm.create_model(self.model_id.replace("timm/", ""), pretrained=True)

    def encode_image(self, image_paths: list[str], batch_size: int = 32) -> np.ndarray:
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model = self.model.to(device)

        all_features = []
        for i in tqdm(range(0, len(image_paths), batch_size), desc="Encoding images"):
            batch_paths = image_paths[i : i + batch_size]
            images = [Image.open(path).convert("RGB") for path in batch_paths]
            preprocessed = torch.stack([self.transform(img) for img in images])
            with torch.inference_mode():
                features = self.model(preprocessed.to(device))
            all_features.append(features.cpu())

        return torch.cat(all_features, dim=0).numpy()



import faiss
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from loguru import logger
from PIL import Image
from rich.console import Console
from rich.table import Table

from .datasets_registry import DatasetRegistry
from .metrics import calculate_retrieval_metrics
from .models.bm25 import BM25sModel
from .models_registry import ModelRegistry


def list_datasets(search: str = ""):
    # Convert wildcard pattern to simple regex-like matching
    search = search.replace("*", "").lower()
    datasets_dict = DatasetRegistry.list()
    filtered_datasets = {
        name: desc for name, desc in datasets_dict.items() if search in name.lower()
    }

    # Create and print table
    table = Table(title="Available Datasets")
    table.add_column("Dataset Name", style="cyan")
    table.add_column("Description", style="magenta")

    for name, description in filtered_datasets.items():
        table.add_row(name, description or "No description available")

    console = Console()
    console.print(table)

    # return datasets


def list_models(search: str = "") -> dict:
    # Convert wildcard pattern to simple regex-like matching
    search = search.replace("*", "").lower()
    # Get filtered models
    models = {
        model_id: model_input
        for model_id, model_input in ModelRegistry.list().items()
        if search in model_id.lower()
    }

    # Create and print table
    table = Table(title="Available Models")
    table.add_column("Model ID", style="cyan")
    table.add_column("Model Input", style="magenta")

    for model_id, input_type in models.items():
        table.add_row(model_id, input_type)

    console = Console()
    console.print(table)

    # return models


def load_dataset(name: str | pd.DataFrame):
    if isinstance(name, pd.DataFrame):
        return name
    dataset_class = DatasetRegistry.get(name)
    return dataset_class.get_dataset()


def load_model(model_id: str):
    model_class = ModelRegistry.get(model_id)
    return model_class(model_id=model_id)


def run_benchmark_bm25(dataset: str, top_k: int = 10):
    logger.info("Running BM25 retrieval benchmark")
    bm25_model = BM25sModel()
    dataset = load_dataset(dataset)

    logger.info("Tokenizing corpus")
    corpus = dataset["caption"].tolist()
    bm25_model.tokenize_text(corpus)

    # Get labels for evaluation
    image_ids = dataset.image_id.tolist()
    image_ids = np.array(image_ids)
    labels = dataset.loc[(dataset.image_id.isin(image_ids))].name.to_numpy()

    logger.info("Performing retrieval")
    retrieved_ids = bm25_model.retrieve(corpus, top_k=top_k)

    logger.info("Calculating metrics")
    eval_metrics_results = calculate_retrieval_metrics(labels, retrieved_ids, top_k)

    # Create results DataFrame for visualization
    results_data = []
    for idx, retrieved in enumerate(retrieved_ids):
        query_name = dataset.iloc[idx]["name"]
        ground_truth_matches = dataset[
            (dataset["name"] == query_name)
            & (dataset["image_id"] != dataset.iloc[idx]["image_id"])
        ]

        query_row = {
            "query_id": dataset.iloc[idx]["image_id"],
            "query_path": dataset.iloc[idx]["image_path"],
            "query_caption": dataset.iloc[idx]["caption"],
            "query_name": dataset.iloc[idx]["name"],
            "retrieved_ids": [dataset.iloc[i]["image_id"] for i in retrieved],
            "retrieved_paths": [dataset.iloc[i]["image_path"] for i in retrieved],
            "retrieved_captions": [dataset.iloc[i]["caption"] for i in retrieved],
            "retrieved_names": [dataset.iloc[i]["name"] for i in retrieved],
            "is_correct": [labels[i] == labels[idx] for i in retrieved],
            "ground_truth_ids": ground_truth_matches["image_id"].tolist(),
            "ground_truth_paths": ground_truth_matches["image_path"].tolist(),
            "ground_truth_captions": ground_truth_matches["caption"].tolist(),
        }
        results_data.append(query_row)

    results_df = pd.DataFrame(results_data)

    return eval_metrics_results, results_df


def run_benchmark(
    dataset: str | pd.DataFrame,
    model_id: str,
    mode: str = "image-to-image",  # Can be "image-to-image", "text-to-text", "text-to-image", or "image-to-text"
    top_k: int = 10,
):
    """
    Run retrieval benchmark on a dataset

    Args:
        dataset_name: Name of the dataset to use or a pandas DataFrame containing the dataset
        model_id: ID of the model to use
        mode: Type of retrieval ("image-to-image", "text-to-text", "text-to-image", or "image-to-text")
        top_k: Number of top results to retrieve (will retrieve top_k + 1 to account for self-matches)
    """
    dataset = load_dataset(dataset)

    # TODO: Dataset should contain columns ['image_id', 'file_name', 'image_path', 'caption', 'name']
    model = load_model(model_id)
    model_info = ModelRegistry.get_model_info(model_id)

    logger.info(f"Evaluating embeddings with parameter top_k: {top_k}")
    image_ids = dataset.image_id.tolist()
    image_ids = np.array(image_ids)
    labels = dataset.loc[(dataset.image_id.isin(image_ids))].name.to_numpy()

    # Encode embeddings based on mode
    if mode.endswith("image"):
        logger.info(f"Encoding database images for {model_id}")
        embeddings = model.encode_image(dataset["image_path"].tolist())
    else:
        logger.info(f"Encoding database text for {model_id}")
        embeddings = model.encode_text(dataset["caption"].tolist())

    if (
        mode.startswith("text")
        and mode.endswith("image")
        or mode.startswith("image")
        and mode.endswith("text")
    ):
        if mode.startswith("image"):
            logger.info(f"Encoding query images for {model_id}")
            query_embeddings = model.encode_image(dataset["image_path"].tolist())
        else:
            logger.info(f"Encoding query text for {model_id}")
            query_embeddings = model.encode_text(dataset["caption"].tolist())
    else:
        query_embeddings = embeddings

    # Create FAISS index
    logger.info("Creating indices for vector search...")
    embedding_dimension = embeddings.shape[1]
    index = faiss.IndexIDMap(faiss.IndexFlatIP(embedding_dimension))
    faiss.normalize_L2(embeddings)
    index.add_with_ids(embeddings, np.arange(len(embeddings)))

    # Search
    logger.info("Performing vector search...")
    faiss.normalize_L2(query_embeddings)
    _, retrieved_ids = index.search(
        query_embeddings, k=top_k + 1
    )  # +1 to account for self-matches

    # Filter self matches
    filtered_retrieved_ids = []
    for idx, row in enumerate(retrieved_ids):
        filtered_row = [x for x in row if x != idx]
        if len(filtered_row) != top_k:
            filtered_row = filtered_row[:top_k]
        filtered_retrieved_ids.append(filtered_row)
    filtered_retrieved_ids = np.array(filtered_retrieved_ids)

    # Calculate metrics
    eval_metrics_results = calculate_retrieval_metrics(
        labels, filtered_retrieved_ids, top_k
    )

    # Create results DataFrame (keeping this for visualization purposes)
    results_data = []
    for idx, retrieved in enumerate(filtered_retrieved_ids):
        # Get all ground truth matches for this query (excluding self)
        query_name = dataset.iloc[idx]["name"]
        ground_truth_matches = dataset[
            (dataset["name"] == query_name)
            & (dataset["image_id"] != dataset.iloc[idx]["image_id"])
        ]

        query_row = {
            "query_id": dataset.iloc[idx]["image_id"],
            "query_path": dataset.iloc[idx]["image_path"],
            "query_caption": dataset.iloc[idx]["caption"],
            "query_name": dataset.iloc[idx]["name"],
            "retrieved_ids": [dataset.iloc[i]["image_id"] for i in retrieved],
            "retrieved_paths": [dataset.iloc[i]["image_path"] for i in retrieved],
            "retrieved_captions": [dataset.iloc[i]["caption"] for i in retrieved],
            "retrieved_names": [dataset.iloc[i]["name"] for i in retrieved],
            "is_correct": [labels[i] == labels[idx] for i in retrieved],
            # Add ground truth information
            "ground_truth_ids": ground_truth_matches["image_id"].tolist(),
            "ground_truth_paths": ground_truth_matches["image_path"].tolist(),
            "ground_truth_captions": ground_truth_matches["caption"].tolist(),
        }
        results_data.append(query_row)

    results_df = pd.DataFrame(results_data)

    return eval_metrics_results, results_df


def visualize_retrieval(
    results_df: pd.DataFrame,
    mode: str | None = None,
    num_queries: int = 5,
    seed: int = 42,
):
    """
    Visualize retrieval results from the benchmark results DataFrame

    Args:
        results_df: DataFrame containing retrieval results from run_benchmark
        mode: Type of retrieval ("image-to-image", "text-to-text", "text-to-image", "image-to-text")
              If None, shows both image and caption for queries and results
        num_queries: Number of random queries to visualize
    """
    np.random.seed(seed)
    # Select random queries
    query_indices = np.random.choice(len(results_df), num_queries, replace=False)

    for query_idx in query_indices:
        query_row = results_df.iloc[query_idx]
        retrieved_paths = query_row["retrieved_paths"]
        retrieved_captions = query_row["retrieved_captions"]
        retrieved_ids = query_row["retrieved_ids"]

        # Always filter out self-matches
        mask = [rid != query_row["query_id"] for rid in retrieved_ids]
        retrieved_paths = [p for p, m in zip(retrieved_paths, mask) if m]
        retrieved_captions = [c for c, m in zip(retrieved_captions, mask) if m]
        retrieved_ids = [i for i, m in zip(retrieved_ids, mask) if m]

        # Limit to 10 results (2 rows of 5)
        max_results = 10
        retrieved_paths = retrieved_paths[:max_results]
        retrieved_captions = retrieved_captions[:max_results]
        retrieved_ids = retrieved_ids[:max_results]
        top_k = len(retrieved_paths)

        plt.figure(figsize=(20, 12))

        plt.subplot(3, 1, 1)
        if mode is None or mode.startswith("image"):
            query_img = Image.open(query_row["query_path"])
            plt.imshow(query_img)
            if mode is None:
                plt.title(
                    f'Query ID: {query_row["query_id"]}\n{query_row["query_caption"][:50]}...',
                    fontsize=10,
                )
            else:
                plt.title(f'Query ID: {query_row["query_id"]}', fontsize=10)
        else:  # text-only mode
            plt.text(
                0.5,
                0.5,
                query_row["query_caption"],
                horizontalalignment="center",
                verticalalignment="center",
                wrap=True,
                fontsize=12,
            )
            plt.title(f'Query ID: {query_row["query_id"]}', fontsize=10)
        plt.axis("off")

        # Retrieved results visualization in 2 rows
        for i in range(top_k):
            row = 1 if i < 5 else 2  # First 5 in row 1, next 5 in row 2
            col = i % 5  # Column position within row
            plt.subplot(3, 5, 5 * row + col + 1)  # Adjusted subplot positioning
            if mode is None or mode.endswith("image"):
                retrieved_img = Image.open(retrieved_paths[i])
                plt.imshow(retrieved_img)
                if mode is None:
                    plt.title(
                        f"Match {i+1} (ID: {retrieved_ids[i]})\n{retrieved_captions[i][:50]}...",
                        fontsize=8,
                    )
                else:
                    plt.title(f"Match {i+1} (ID: {retrieved_ids[i]})", fontsize=8)
            else:  # text-only mode
                plt.text(
                    0.5,
                    0.5,
                    retrieved_captions[i],
                    horizontalalignment="center",
                    verticalalignment="center",
                    wrap=True,
                    fontsize=8,
                )
                plt.title(f"Match {i+1} (ID: {retrieved_ids[i]})", fontsize=8)
            plt.axis("off")

        plt.tight_layout(h_pad=2, w_pad=1)
        plt.show()


def visualize_ground_truth(
    dataset: str,
    mode: str | None = None,
    num_queries: int = 5,
    seed: int = 42,
):
    """
    Visualize ground truth matches from the dataset

    Args:
        dataset: Dataset name or DataFrame containing the dataset
        mode: Type of visualization ("image-to-image", "text-to-text", "text-to-image", "image-to-text")
              If None, shows both image and caption for queries and results
        num_queries: Number of random queries to visualize
        seed: Random seed for reproducibility
    """
    np.random.seed(seed)
    dataset = load_dataset(dataset)

    # Select random queries
    query_indices = np.random.choice(len(dataset), num_queries, replace=False)

    for query_idx in query_indices:
        query_row = dataset.iloc[query_idx]
        query_name = query_row["name"]

        # Get all matches (excluding self)
        matches = dataset[
            (dataset["name"] == query_name)
            & (dataset["image_id"] != query_row["image_id"])
        ]

        # Limit to 10 results (2 rows of 5)
        max_results = 10
        matches = matches.head(max_results)
        top_k = len(matches)

        plt.figure(figsize=(20, 12))

        # Query visualization
        plt.subplot(3, 1, 1)
        if mode is None or mode.startswith("image"):
            query_img = Image.open(query_row["image_path"])
            plt.imshow(query_img)
            if mode is None:
                plt.title(
                    f'Query ID: {query_row["image_id"]}\n{query_row["caption"][:50]}...',
                    fontsize=10,
                )
            else:
                plt.title(f'Query ID: {query_row["image_id"]}', fontsize=10)
        else:  # text-only mode
            plt.text(
                0.5,
                0.5,
                query_row["caption"],
                horizontalalignment="center",
                verticalalignment="center",
                wrap=True,
                fontsize=12,
            )
            plt.title(f'Query ID: {query_row["image_id"]}', fontsize=10)
        plt.axis("off")

        # Ground truth matches visualization in 2 rows
        for i, (_, match) in enumerate(matches.iterrows()):
            row = 1 if i < 5 else 2  # First 5 in row 1, next 5 in row 2
            col = i % 5  # Column position within row
            plt.subplot(3, 5, 5 * row + col + 1)  # Adjusted subplot positioning

            if mode is None or mode.endswith("image"):
                match_img = Image.open(match["image_path"])
                plt.imshow(match_img)
                if mode is None:
                    plt.title(
                        f"Match {i+1} (ID: {match['image_id']})\n{match['caption'][:50]}...",
                        fontsize=8,
                    )
                else:
                    plt.title(f"Match {i+1} (ID: {match['image_id']})", fontsize=8)
            else:  # text-only mode
                plt.text(
                    0.5,
                    0.5,
                    match["caption"],
                    horizontalalignment="center",
                    verticalalignment="center",
                    wrap=True,
                    fontsize=8,
                )
                plt.title(f"Match {i+1} (ID: {match['image_id']})", fontsize=8)
            plt.axis("off")

        plt.tight_layout(h_pad=2, w_pad=1)
        plt.show()



from abc import ABC, abstractmethod


class MultiModalModel(ABC):
    """
    A model that takes both text and image as input.
    """

    def __init__(self, model_id: str):
        self.model_id = model_id

    @abstractmethod
    def encode_image(self, image_path: str):
        pass

    @abstractmethod
    def encode_text(self, text: str):
        pass


class TextModel(ABC):
    """
    A model that only takes text as input.
    """

    def __init__(self, model_id: str):
        self.model_id = model_id

    @abstractmethod
    def encode_text(self, text: str):
        pass


class ImageModel(ABC):
    """
    A model that only takes image as input.
    """

    def __init__(self, model_id: str):
        self.model_id = model_id

    @abstractmethod
    def encode_image(self, image_path: str):
        pass



import numpy as np
from sentence_transformers import SentenceTransformer
from tqdm.auto import tqdm

from xretrieval.models.base import TextModel
from xretrieval.models_registry import ModelRegistry


@ModelRegistry.register("sentence-transformers/all-MiniLM-L6-v2", model_input="text")
@ModelRegistry.register(
    "sentence-transformers/multi-qa-MiniLM-L6-cos-v1", model_input="text"
)
@ModelRegistry.register("sentence-transformers/all-mpnet-base-v2", model_input="text")
@ModelRegistry.register(
    "sentence-transformers/multi-qa-mpnet-base-dot-v1", model_input="text"
)
@ModelRegistry.register(
    "sentence-transformers/all-distilroberta-v1", model_input="text"
)
@ModelRegistry.register("sentence-transformers/all-MiniLM-L12-v2", model_input="text")
@ModelRegistry.register(
    "sentence-transformers/multi-qa-distilbert-cos-v1", model_input="text"
)
@ModelRegistry.register(
    "sentence-transformers/paraphrase-albert-small-v2", model_input="text"
)
@ModelRegistry.register(
    "sentence-transformers/paraphrase-MiniLM-L3-v2", model_input="text"
)
class SentenceTransformerModel(TextModel):
    def __init__(self, model_id: str):
        self.model_id = model_id
        self.model = self.load_model()

    def load_model(self):
        return SentenceTransformer(self.model_id)

    def encode_text(self, text: list[str], batch_size: int = 32):
        all_features = []
        for i in tqdm(range(0, len(text), batch_size), desc="Encoding text"):
            batch_text = text[i : i + batch_size]
            features = self.model.encode(batch_text)
            all_features.append(features)
        return np.concatenate(all_features, axis=0)
